//---------------------------------------------------------------------------
// By Cozmo, 2019-2021
//
// ANNAWeapon aims to remove lead-aiming from the game by using partial 
// clientside hit detection. When you shoot someone on your client, the
// server will recreate that shot with up-to-date info to confirm it.
//
// Client and server firing functions are also now much more distinct. This 
// allows for quicker feedback and improved security against cheats.
//
// 06/2020 update: Vanilla NPCs can now fire these weapons normally in MP
//---------------------------------------------------------------------------
class ANNAWeapon extends DeusExWeapon abstract;

//#exec TEXTURE IMPORT NAME=DX_HitMarker FILE=Textures\DX_HitMarker.pcx GROUP=UI
#exec AUDIO IMPORT FILE="Sounds\QFeedBack.wav" NAME="QFeedBack" GROUP="Generic"

var float nextShotTime;

//weapon attributes:
var float bringUpTime;
var float putAwayTime;
var int numShots;
var int lowClipCount;
var float burstDuration;
var float recoilDuration;
var float gunAccuracy;
var float atkDelayTime;
var float altAmmoDelayTime; //from equipping wep to when you can fire with alt ammo (prevents rapid firing)
var float projectileSpeed;  //todo? maybe not
var float deleteAfterFireTime;
var float delWepTime;
var bool bMissingAnims; //true if the weapon model lacks animations (like fire extinguisher)

//other:
var Actor qTarget[20];
var vector qOffset[20];
var float qShotTime;
var float stopBurstTime;
var float finishAtkTime;
var bool bBursting;
var bool bH2HAttacking;
var int curNumShot;
var bool bIsReloading;
var bool bWasZoomed;
var float recoilUntil;
var Sound HitFeedback;

const DSOC_FIREANIMS = 1;
const DSOC_DEMOSOUND = 2;

replication
{
	reliable if(Role < ROLE_Authority)
		FireOnServer, ServerStartH2HAttack; //ServerFinishH2HAttack;
		
	reliable if(Role == ROLE_Authority)
		BringUpClient, SpawnEffectsClient, SetClientCanFireTime;
		
	//reliable if((Role == ROLE_Authority) && (bNetOwner == false))
	//	PlayWepFiring;
}

//Cozmo: I know I shouldn't have two functions that do the same thing, but I can't make it work simultaneously on serv/client for some reason
function BringUp()
{
	if(AmmoIsDefault() == false) //no surprise-shelling or rapid shelling
		nextShotTime = Level.TimeSeconds + altAmmoDelayTime;
	else
	{
		if(nextShotTime > Level.TimeSeconds + 10)	//hack!!! (for switching mid-reload)
			nextShotTime -= 10000;
		
		if(nextShotTime < Level.TimeSeconds + bringUpTime)
			nextShotTime = Level.TimeSeconds + bringUpTime;
	}
		
	bIsReloading = false;
	if(bMissingAnims == false) PlayAnim('Select',1.0,0);
	Owner.PlaySound(SelectSound, SLOT_Misc, Pawn(Owner).SoundDampening); //todo: replace with PlayOwnedSound()?
	ClipCount = 0;
	bZoomed = False;
	bWasZoomed = false;
	
	//serverside stuff
	BringUpClient();
	Super.BringUp();
}

simulated function BringUpClient() //in DXMP, weapon switching isn't clientside whatsoever. Waiting for serv causes lag, but changing that is probably a TON of work
{
	if(AmmoIsDefault() == false) //no surprise-shelling or rapid shelling
		nextShotTime = Level.TimeSeconds + altAmmoDelayTime;
	else
	{
		if(nextShotTime > Level.TimeSeconds + 10)	//hack!!! (for switching mid-reload)
			nextShotTime -= 10000;

		if(nextShotTime < Level.TimeSeconds + bringUpTime)
			nextShotTime = Level.TimeSeconds + bringUpTime;
	}

	bIsReloading = false;
	if(bMissingAnims == false) PlayAnim('Select',1.0,0);
	Owner.PlaySound(SelectSound, SLOT_Misc, Pawn(Owner).SoundDampening);
	ClipCount = 0;
	bZoomed = False;
	bWasZoomed = false;
}

function bool PutDown() //I'm not sure this function is even being called?
{
	//broadcastmessage("wep putdown!"); //I don't think it is...
	GotoState('DownWeapon');
	bWasZoomed = false;
	return True;
}

function TweenSelect()
{
	if(bMissingAnims == false) TweenAnim('Select',0.001);
}

function PlaySelect()
{
	if(bMissingAnims == false) PlayAnim('Select',1.0,0.0);
	Owner.PlaySound(SelectSound, SLOT_Misc, Pawn(Owner).SoundDampening);	
}

//state DownWeapon
//{
//ignores Fire, AltFire;
//
//	function bool PutDown()
//	{
//		// alert NPCs that I'm putting away my gun
//		AIEndEvent('WeaponDrawn', EAITYPE_Visual);
//		return Super.PutDown();
//	}
//
//Begin:
//	ScopeOff();
//	LaserOff();
//
//	if (( Level.NetMode == NM_DedicatedServer ) || ((Level.NetMode == NM_ListenServer) && Owner.IsA('DeusExPlayer') && !DeusExPlayer(Owner).PlayerIsListenClient()))
//		ClientDownWeapon();
//
//	TweenDown();
//	FinishAnim();
//
//	if ( Level.NetMode != NM_Standalone )
//	{
//		ClipCount = 0;	// Auto-reload in multiplayer (when putting away)
//	}
//	bOnlyOwnerSee = false;
//	if (Pawn(Owner) != None)
//		Pawn(Owner).ChangedWeapon();
//}

simulated function SwapMuzzleFlashTexture()
{
/*
   if (!bHasMuzzleFlash)
      return;
	if (FRand() < 0.5)
		MultiSkins[2] = Texture'FlatFXTex34';
	else
		MultiSkins[2] = Texture'FlatFXTex37';

	MuzzleFlashLight();
	SetTimer(0.1, False);
	*/
}

simulated function EraseMuzzleFlashTexture()
{
	MultiSkins[2] = None;
}

simulated function Tick(float deltaTime)
{
	local DeusExPlayer dxp, player;
	local Pawn pwn;
	local int i;
	local Vector HitLocation, HitNormal, StartTrace, EndTrace, X, Y, Z, offset;
	local Rotator rot;
	local actor target, RealTarget;
	local float beepspeed, recoil;
	
	if(deleteAfterFireTime > 0 && delWepTime != 0)
	{
		if(Level.TimeSeconds >= delWepTime)
		{
			Destroy();
		}
	}
	
	if(Owner == None)
		return;
	
	if(Level.NetMode != NM_DedicatedServer)
	{
		dxp = DeusExPlayer(Owner);
		
		if(dxp != None)
		{
			if(dxp.Weapon == Self && ((dxp.bFire != 0 ) || (Level.TimeSeconds < stopBurstTime))) //bFire is "button bFire"
			{																					 //using ClientFire() breaks autofire. Not worth debugging
			//	PlayWepFiring(); //calling this inside FoC() creates lag?
				FireOnClient();
			}
			
			if(dxp.Weapon == Self)
			{
				if(bH2HAttacking && Level.TimeSeconds >= finishAtkTime)
				{
					bH2HAttacking = false;
					
					if(ProjectileClass == None) //if melee
					{
						GetAxes(Pawn(Owner).ViewRotation, X, Y, Z);
						StartTrace = Pawn(Owner).Location;
						EndTrace = StartTrace + mpMaxRange * X;
						StartTrace.Z += Pawn(Owner).BaseEyeHeight;
						EndTrace.Z += Pawn(Owner).BaseEyeHeight;

						target = Trace(HitLocation, HitNormal, EndTrace, StartTrace, True);
						
						if(target != None)
							offset = HitLocation - target.Location;
					}

					//check target type, and play the appropriate sound on client (play it later on server as owned sound)
					if(target != None)
					{
						if (target.IsA('DeusExDecoration'))
							PlaySound(Misc3Sound, SLOT_None,,, 1024);
						else if (target.IsA('Pawn'))
							PlaySound(Misc1Sound, SLOT_None,,, 1024);
						else if (target.IsA('BreakableGlass'))
							PlaySound(sound'GlassHit1', SLOT_None,,, 1024);
						else if (GetWallMaterial(HitLocation, HitNormal) == 'Glass')
							PlaySound(sound'BulletProofHit', SLOT_None,,, 1024);
						else if(target != none)
							PlaySound(Misc2Sound, SLOT_None,,, 1024);
					}
					
						//ServerFinishH2HAttack(target);
					FireOnServer(target, offset);
				}
			}
		}
	}
	
	if(Role == ROLE_Authority)
	{
		//if(Owner.IsA('ScriptedPawn') && burstDuration > 0 && Level.TimeSeconds < stopBurstTime)
		if(ScriptedPawn(Owner) != None && burstDuration > 0 && Level.TimeSeconds < stopBurstTime)
			FireOnServer();
		
		if(qShotTime != 0)
		{
			if(Level.TimeSeconds >= qShotTime)
			{
				for(i=0; i < numShots; i++)
				{
					//broadcastmessage("debug: Queued shot called. i: "$i$", curNumShot: "$curNumShot);
					FireOnServer(qTarget[i], qOffset[i]);
					qTarget[i] = none;
				}
			
				//broadcastmessage("debug: trying to fire shot from queue");
				//FireOnServer(qTarget, qOffset);
				//qTarget = None;
				//qOffset *= 0;
				qShotTime = 0;
			}
		}
		
		if((Level.TimeSeconds > finishAtkTime + 1.5) && (bH2HAttacking == true)) //failsafe (don't allow players to hold a h2h delay indefinitely)
			bH2HAttacking = false;
	}
	
	// GEP lock-on stuff
	
	player = DeusExPlayer(Owner);
	pwn = Pawn(Owner);
	
	if(pwn == None)
	{
		LockMode = LOCK_None;
		MaintainLockTimer = 0;
		LockTarget = None;
		LockTimer = 0;
		return;
	}

	if (pwn.Weapon != self)
	{
		LockMode = LOCK_None;
		MaintainLockTimer = 0;
		LockTarget = None;
		LockTimer = 0;
		return;
	}

	// all this should only happen IF you have ammo loaded
	if (ClipCount < ReloadCount)
	{
		// check for LAM or other placed mine placement
		//if (bHandToHand && (ProjectileClass != None) && (!Self.IsA('WeaponShuriken')))

		SoundTimer += deltaTime;

		if((Level.Netmode == NM_Standalone) || ((Player != None) && (Player.PlayerIsClient())))
		{
			if (bCanTrack)
			{
				Target = AcquireTarget();
				RealTarget = Target;
            
				// calculate the range
				if(Target != None)
					TargetRange = Abs(VSize(Target.Location - Location));
            
				// update our timers
				//SoundTimer += deltaTime;
				MaintainLockTimer -= deltaTime;
            
				// check target and range info to see what our mode is
				if((Target == None) || IsInState('Reload'))
				{
					if (MaintainLockTimer <= 0)
					{				
						SetLockMode(LOCK_None);
						MaintainLockTimer = 0;
						LockTarget = None;
					}
					else if (LockMode == LOCK_Locked)
					{
					  Target = LockTarget;
					}
				}
				else if((Target != LockTarget) && (Target.IsA('Pawn')) && (LockMode == LOCK_Locked))
				{
					SetLockMode(LOCK_None);
					LockTarget = None;
				}
				else if(!Target.IsA('Pawn'))
				{
					if (MaintainLockTimer <=0 )
						SetLockMode(LOCK_Invalid);
				}
				else if((Target.IsA('DeusExPlayer')) && (Target.Style == STY_Translucent)) //Cozmo: I've tried adding Radar Trans too, but lock-on is clientside and I don't think client has that info
				{
					//DEUS_EX AMSD Don't allow locks on cloaked targets.
					SetLockMode(LOCK_Invalid);
				}
				else if((Target.IsA('DeusExPlayer')) && (Player.DXGame.IsA('TeamDMGame')) && (TeamDMGame(Player.DXGame).ArePlayersAllied(Player,DeusExPlayer(Target))))
				{
					//DEUS_EX AMSD Don't allow locks on allies.
					SetLockMode(LOCK_Invalid);
				}
				else
				{
					if(TargetRange > mpMaxRange)
					{
						SetLockMode(LOCK_Range);
					}
					else
					{
						// change LockTime based on skill
						// -0.7 = max skill
						// DEUS_EX AMSD Only do weaponskill check here when first checking.
						if(LockTimer == 0)
						{
							//todo: maybe get lockon time from a table instead, like GetWeaponDamage()?
							LockTime = FMax(Default.LockTime + 3.0 * GetWeaponSkill(), 0.0);
							
							if((Level.Netmode != NM_Standalone) && (LockTime < 0.25))
								LockTime = 0.25;
						}

						LockTimer += deltaTime;
						
						if(LockTimer >= LockTime)
							SetLockMode(LOCK_Locked);
						else
							SetLockMode(LOCK_Acquire);
					}
				}

				// act on the lock mode
				switch(LockMode)
				{
					case LOCK_None:
					TargetMessage = msgNone;
					LockTimer -= deltaTime;
					break;

					case LOCK_Invalid:
					TargetMessage = msgLockInvalid;
					LockTimer -= deltaTime;
					break;

					case LOCK_Range:
					TargetMessage = msgLockRange @ Int(TargetRange/16) @ msgRangeUnit;
					LockTimer -= deltaTime;
					break;

					case LOCK_Acquire:
					TargetMessage = msgLockAcquire @ Left(String(LockTime-LockTimer), 4) @ msgTimeUnit;
					beepspeed = FClamp((LockTime - LockTimer) / Default.LockTime, 0.2, 1.0);
					
					if(SoundTimer > beepspeed)
					{
						Owner.PlaySound(TrackingSound, SLOT_None);
						SoundTimer = 0;
					}
					break;

					case LOCK_Locked:
					// If maintaining a lock, or getting a new one, increment maintainlocktimer
					if ((RealTarget != None) && ((RealTarget == LockTarget) || (LockTarget == None)))
					{
						if (Level.NetMode != NM_Standalone)
							MaintainLockTimer = default.MaintainLockTimer;
						else
							MaintainLockTimer = 0;
							
						LockTarget = Target;
					}
					
					TargetMessage = msgLockLocked @ Int(TargetRange/16) @ msgRangeUnit;
					// DEUS_EX AMSD Moved out so server can play it so that client knows for sure when locked.
					/*if (SoundTimer > 0.1)
					{
					Owner.PlaySound(LockedSound, SLOT_None);
					SoundTimer = 0;
					}*/
					break;
				}
			}
			else
			{
				LockMode = LOCK_None;
				TargetMessage = msgNone;
				LockTimer = 0;
				MaintainLockTimer = 0;
				LockTarget = None;
			}

			if(LockTimer < 0)
				LockTimer = 0;
		}
	}
	else
	{
		LockMode = LOCK_None;
		TargetMessage=msgNone;
		MaintainLockTimer = 0;
		LockTarget = None;
		LockTimer = 0;
	}

	if((LockMode == LOCK_Locked) && (SoundTimer > 0.1) && (Role == ROLE_Authority))
	{
		PlayLockSound();
		SoundTimer = 0;
	}
	
	if(Level.TimeSeconds < recoilUntil && default.recoilDuration > 0)
	{
		recoil = GetWeaponRecoil();
		
		player.ViewRotation.Yaw += deltaTime * (Rand(4096) - 2048) * recoil; //note: Pitch = vertical, Yaw = horizontal. Roll = not needed here
		player.ViewRotation.Pitch += deltaTime * (Rand(4096) + 4096) * recoil;
		
		if ((player.ViewRotation.Pitch > 16384) && (player.ViewRotation.Pitch < 32768))
			player.ViewRotation.Pitch = 16384;
	}
	
//	Super(Weapon).Tick(deltaTime);
}

function ServerStartH2HAttack()
{
	if(Level.TimeSeconds >= finishAtkTime)
	{
		finishAtkTime = Level.TimeSeconds + atkDelayTime;
		bH2HAttacking = true;
		Pawn(Owner).PlayFiring();
		dsoc(DSOC_FIREANIMS); //move upwards for hand-to-hand weapons?
		PlayOwnedSound(FireSound, SLOT_None, TransientSoundVolume);
	}
}

simulated function StartAtkAnim()
{
	if(bMissingAnims == false) PlayAnim('Attack');
}


// ----------------------------------------------------------------------
// FireOnClient()
//
// Spawn some instant weapon feedback, and send our target + vector of
// where target was hit to the server
// ----------------------------------------------------------------------
simulated function FireOnClient()
{
	local Actor myTarget, tracer;
	local Vector HitLocation, HitNormal, StartTrace, EndTrace, X, Y, Z, offset;
	local Rotator rot;
	local Pawn dxp;
	local int i, shotRot, spreadGap, spreadMargin;
	local float mult;
	local Spark spark;

	dxp = Pawn(Owner);
	
	if((dxp.Weapon != Self) || (Level.NetMode == NM_DedicatedServer)) // || (IsInState('DownWeapon'))) //doesn't work for client :(
		return;
		
	// check for surrounding environment
	if((EnviroEffective == ENVEFF_Air) || (EnviroEffective == ENVEFF_Vacuum) || (EnviroEffective == ENVEFF_AirVacuum))
	{
		if(Region.Zone.bWaterZone)
			return;
	}
		
	if(Level.TimeSeconds >= nextShotTime) 
	{
		if((ClipCount < ReloadCount) || (ReloadCount == 0))
		{					
			//if(AmmoType != None)
			if(AmmoName != Class'DeusEx.AmmoNone') //lol
			{
				AmmoType.SimUseAmmo();
				
				if(AmmoType.AmmoAmount < 1)
					return;
			}
					
			if(bHandToHand)
			{
				//start attack if we're not already attacking 
				
				if(finishAtkTime < Level.TimeSeconds) // <-------------- h2h stuff is here
				{
					StartAtkAnim();
					bH2HAttacking = true;
					//play sound atk
					finishAtkTime = Level.TimeSeconds + atkDelayTime;
					ServerStartH2HAttack();
				}
			}
			else
			{
				spreadGap = GetWeaponSpread();
						
				for(i = 0; i < numShots; i++)
				{
					if(bInstantHit)
					{
						spreadMargin = GetWeaponSpreadMargin();
						spreadMargin = ((frand() * spreadMargin) - (spreadMargin)); //GetWeaponSpreadMargin() can't return a float like 0.5 so nothing fancy here
						
						//get target
						shotRot = Rand(360); //note: (x,y) = (12*sin(shotRot), 12*cos(shotRot))
						GetAxes(dxp.ViewRotation, X, Y, Z);
						StartTrace = dxp.Location;
						EndTrace = StartTrace + (((i*spreadGap)+spreadMargin)*Sin(shotRot))*Y + (((i*spreadGap)+spreadMargin)*Cos(shotRot))*Z;
						EndTrace += mpMaxRange * X;
						StartTrace.Z += dxp.BaseEyeHeight;
						EndTrace.Z += dxp.BaseEyeHeight;
						
						myTarget = Trace(HitLocation, HitNormal, EndTrace, StartTrace, True);
						
						if(AmmoName != class'AmmoBattery')
						{
							//spawn tracer
							StartTrace += 10 * X;
							rot = Rotator(EndTrace - StartTrace);
							tracer = Spawn(Class'TracerClient', Owner,, StartTrace, rot);
							tracer.remoterole = ROLE_None;
						}
									
						if(myTarget != None)
						{
							offset = HitLocation - myTarget.Location; // << myTarget.Rotation; //rotation isn't actually needed from client..
						
							if((myTarget == Level) || (myTarget.IsA('Mover')))
								SpawnEffectsClient(HitLocation, HitNormal, myTarget, 20);
							
							if(myTarget.IsA('DeusExDecoration'))
							{
								if(DeusExDecoration(myTarget).bInvincible == true || DeusExDecoration(myTarget).default.HitPoints > 2000) //spawn sparks on invincible deco
								{
									spark = spawn(class'Spark', Owner,, HitLocation, rot);
									spark.RemoteRole = ROLE_None;
								}
							}
							else if(myTarget.IsA('DeusExPlayer'))
							{
								mult = 1.2;
								
								if(numShots > 1) //shotguns hit can hit multiple times and be very loud
									mult = 0.7;
									
								if(TeamDMGame(Level.Game) != None) 
								{
									if(Pawn(myTarget).PlayerReplicationInfo.Team != Pawn(Owner).PlayerReplicationInfo.Team)
									{
										if(Pawn(myTarget).Health > 0)
											PlaySound(HitFeedback, SLOT_None, TransientSoundVolume*mult,, 2048); 
									}
								}
								else
								{
									PlaySound(HitFeedback, SLOT_None, TransientSoundVolume*mult,, 2048); 
								}
							}
						}
					}
					
					FireOnServer(myTarget, offset*1000, LockTarget); //scaling offset up for replication accuracy
				}
			
				 //wep FX stuff:	
				if(burstDuration > 0)
				{
					if(Level.TimeSeconds >= stopBurstTime)
					{
						if(bMissingAnims == false)
						{
							if(bHandToHand)
								PlayAnim('Attack',1.0);
							else
								PlayAnim('Shoot',1.0);
						}
										
						PlaySound(FireSound, SLOT_None, TransientSoundVolume,, 2048 ); 
						//PlaySound(Sound snd, ESoundSlot Slot, float Volume, float Radius )
						stopBurstTime = Level.TimeSeconds + burstDuration;
						recoilUntil = Level.TimeSeconds + default.recoilDuration;
						
						if (bHasMuzzleFlash)
						{
							MuzzleFlashLight();
							SetTimer(0.5, False); //AR looks very spazzy doing lighting for every shot - even if it's correct...
						}
					}
				}
				else
				{
					if(bMissingAnims == false) 
					{
						if(bHandToHand)
							PlayAnim('Attack',1.0);
						else
							PlayAnim('Shoot',1.0);
					}
						
					if (bHasMuzzleFlash)
					{
						MuzzleFlashLight();
						SetTimer(0.1, False);
					}
							
					//PlayAnim('Shoot', 0.8);
					PlaySound(FireSound, SLOT_None, TransientSoundVolume,, 2048 ); //(Sound snd, ESoundSlot Slot, float Volume, float Radius )
					//PlaySimSound(FireSound, SLOT_None, TransientSoundVolume, 2048 ); //(Sound snd, ESoundSlot Slot, float Volume, float Radius )
					recoilUntil = Level.TimeSeconds + default.recoilDuration;
				}
			}
		}
	//	else
	//	{
	//		PlaySound(sound'DryFire', SLOT_None);
	//		//dryfire sound (client, then on server)
	//		//reload
	//	}
		
		dxp.PlayFiring();
		
		//Pawn(Owner).ViewRotation.Pitch += 200; //recoil test. Pitch = vertical, yaw = horizontal. roll = unwanted here (just this line made it feel very jarring. use in tick() like default)
		
		if(AmmoIsDefault() == true)
			nextShotTime = Level.TimeSeconds + shotTime;
		else
			nextShotTime = level.TimeSeconds + 1.5;
	}
}

simulated function int GetWeaponSpread() //this is unintuitive, but only use this for shotguns. Use below function for things like assault rifles
{
	return 40; //default value only
}

simulated function int GetWeaponSpreadMargin()
{
	return 0; //default value only 
}

function bool CheckForAmmo(optional bool bIgnoreClip)
{
	if(Pawn(Owner).IsA('DeusExPlayer')) //for now, bots don't need to worry about reloads or ammo
	{
		//if(AmmoType != None)
		if(AmmoName != Class'DeusEx.AmmoNone')
		{
			if(AmmoType.AmmoAmount > 0)
			{
				if(bIgnoreClip == true)
					return true;
					
				if(ClipCount < ReloadCount)
					return true;
			}
			
			//broadcastmessage("ClipCount: "$ClipCount$", ReloadCount: "$ReloadCount);
			
			//UseAmmo() seems to spawn shellcasings SERVERSIDE sigh
			//if(AmmoType.UseAmmo(1) == false)
			//	return;
		}
		else
		{
			return true;
		}
	}
	else
	{
		return true;
	}
	
	return false;
}

function PlayAppropriateSound(vector HitLocation, vector HitNormal, optional actor target)
{
	if(bHandToHand == false)
		return;
		
	if(target != none)
	{
		if (target.IsA('DeusExDecoration'))
			PlayOwnedSound(Misc3Sound, SLOT_None,,, 1024);
		else if (target.IsA('Pawn'))
			PlayOwnedSound(Misc1Sound, SLOT_None,,, 1024);
		else if (target.IsA('BreakableGlass'))
			PlayOwnedSound(sound'GlassHit1', SLOT_None,,, 1024);
		else if (GetWallMaterial(HitLocation, HitNormal) == 'Glass')
			PlayOwnedSound(sound'BulletProofHit', SLOT_None,,, 1024);
		else
			PlayOwnedSound(Misc2Sound, SLOT_None,,, 1024);
	}
}

// ----------------------------------------------------------------------
// FireOnServer()
//
// NOTE! DryFiring might be awkward for other players, because AmmoAmount is only repl
//	to the weapon owner!!!!
// ----------------------------------------------------------------------

function FireOnServer(optional Actor target, optional Vector offset, optional Actor trackTarget)
{
	local Vector HitLocation, HitNormal, StartTrace, EndTrace, X, Y, Z;
	local Rotator rot;
	local actor a, tfw;
	local Pawn pawn1, dxp;
	local int h2hStatus, curMaxRange;
	local name dmgType;
	
	if(Role == ROLE_Authority) //SERVER ONLY //remove? shouldn't matter if client can speed up calls to this?
	{
		if(Pawn(Owner).Weapon != Self) // || (IsInState('DownWeapon'))) //state doesn't work for client?
			return;
			
		if(Pawn(Owner).Health <= 0 || Pawn(Owner).HealthHead <= 0 || Pawn(Owner).HealthTorso <= 0)
			return;

		if(Level.TimeSeconds >= nextShotTime)
		{
			dmgType = WeaponDamageType();
			curMaxRange = mpMaxRange;
			
			if(ScriptedPawn(Owner) != None)
				curMaxRange += 60;
			
			if(bHandToHand == true)
			{
				curMaxRange += 60; //some extra serverside leniency for melee weps, should hopefully prevent false positives and wrong hitlocs
				
				if(bH2HAttacking == false && DeusExPlayer(Owner) != None) //let NPCs attack instantly
					return;
					
				if(Level.TimeSeconds < finishAtkTime)
				{
					//broadcastmessage("debug: h2h too early, added to queue");
					qShotTime = finishAtkTime; //was nextShotTime
					qTarget[curNumShot] = target;
					qOffset[curNumShot] = offset;
					return;
				}
			}
			
			// check for surrounding environment
			if((EnviroEffective == ENVEFF_Air) || (EnviroEffective == ENVEFF_Vacuum) || (EnviroEffective == ENVEFF_AirVacuum))
			{
				if(Region.Zone.bWaterZone)				
					return;
			}
			
			//PlayWepFiring(); //to all except owner, hopefully
			pawn1 = Pawn(Owner);
			bH2HAttacking = false;
			
			if(curNumShot == 0)
			{
				if(CheckForAmmo() == true)
				{
					if(AmmoName != Class'DeusEx.AmmoNone' && Owner.IsA('DeusExPlayer')) //only players have to worry about ammo atm
					{
						AmmoType.AmmoAmount--;
						ClipCount++;
						
						//broadcastmessage("used ammo. "$ammoType.AmmoAmount$" remains");
						
						if(AmmoType.AmmoAmount < 0)
							AmmoType.AmmoAmount = 0;
							
						if((ClipCount >= ReloadCount) && (!bHandToHand))
						{
							//PlaySound(sound'Switch1Click', SLOT_None, TransientSoundVolume); //when we run out of ammo, play a click //todo: change to a dryfire sound on last shots?
							
							if(!IsInState('DownWeapon')) 
								ReloadAmmo();
						}
							
						//increase click volume!
						//PlaySound (Sound Sound, optional ESoundSlot Slot, optional float Volume, optional bool bNoOverride, optional float Radius, optional float Pitch) [native, final]
					}
				}
				else
				{
					//broadcastmessage("debug: ammo check failed");
					return;
				}
				
				/*  */
				if(burstDuration > 0)
				{
					if(Level.TimeSeconds >= stopBurstTime)
					{
						PlayOwnedSound(FireSound, SLOT_None, TransientSoundVolume);
						dsoc(DSOC_FIREANIMS); //moved here to fix a crash, but this won't correctly muzzleflash etc for all shots?
						dsoc(DSOC_DEMOSOUND);
						stopBurstTime = Level.TimeSeconds + burstDuration;
					}
				}
				else
				{
					if(bHandToHand == False)
					{
						PlayOwnedSound(FireSound, SLOT_None, TransientSoundVolume);
						dsoc(DSOC_FIREANIMS); //move upwards for hand-to-hand weapons?
						dsoc(DSOC_DEMOSOUND);
					}
				}
			}
			
			if(deleteAfterFireTime > 0 && delWepTime == 0)
			{
				delWepTime = Level.TimeSeconds + deleteAfterFireTime;
				//mark this wep for deletion
			}
			
			if(bInstantHit)
			{
				StartTrace = pawn1.Location;
				StartTrace.Z += pawn1.BaseEyeHeight;
					
				if((target != None ) && (target.IsA('Mover') || target == Level))
					target = None;
					
				if(target != none)
				{
					offset *= 0.001; //was scaled up for extra precision because replicating floats rounds them :/
					offset.X *= 0.9; //move offset slightly deeper into the collision cylinder
					offset.Y *= 0.9;
					
					if((offset.z >= target.CollisionHeight) || (offset.z <= -1 * target.CollisionHeight)) //only scale offset vertically if it's too high/low
						offset.Z *= 0.96;
						
					EndTrace = target.Location + offset;
					a = Trace(HitLocation, HitNormal, EndTrace, StartTrace, True);
					
					if(vsize(StartTrace - EndTrace) > curMaxRange) //shot was too far to be possible
						a = None;
					
					if(a != None)
					{
						a.TakeDamage(GetWeaponDamage(target), Pawn(Owner), HitLocation, vect(4,5,1), dmgType);
						SpawnServerTracer(EndTrace, false);
						TrySpawnBlood(a, dmgType, HitLocation, HitNormal);
						PlayAppropriateSound(HitLocation, HitNormal, a);
						
					//	if(a.IsA('Pawn') && !a.IsA('Robot'))
					//		SpawnBlood(HitLocation, HitNormal);
					}
					else //shot with offset failed. Try one last basic trace:
					{
						EndTrace = target.Location;
						a = Trace(HitLocation, HitNormal, EndTrace, StartTrace, True);
							
						if(vsize(StartTrace - EndTrace) > curMaxRange) //shot was too far to be possible
							a = None;
							
						if(a != None)
						{
							a.TakeDamage(GetWeaponDamage(target), Pawn(Owner), HitLocation, vect(4,5,1), dmgType);
							SpawnServerTracer(EndTrace, false);
							TrySpawnBlood(a, dmgType, HitLocation, HitNormal);
							PlayAppropriateSound(HitLocation, HitNormal, a);
						}
					}
				}
				else //has shot either level, mover or none (max range) with hitscan
				{
					if(pawn1.IsA('DeusExPlayer')) //if player, you can't shoot actors unless they're a target - only level & movers (prevents double hitboxes)
					{
						EndTrace = StartTrace + Vector(pawn1.ViewRotation) * curMaxRange;
						a = Trace(HitLocation, HitNormal, EndTrace, StartTrace, False);
						
						if(a != none)
						{
							if((a == Level) || (a.IsA('Mover')))
							{
								a.TakeDamage(GetWeaponDamage(target), Pawn(Owner), HitLocation, vect(0,0,0), dmgType);
								PlayAppropriateSound(HitLocation, HitNormal, a);
							}
						}
						
						SpawnServerTracer(EndTrace, true);
					}
					else
					{
						//---------------------------------------------------
						// Owner is NPC. Use vanilla-style firing code here!
						//---------------------------------------------------
						GetAxes(pawn1.ViewRotation, X, Y, Z);
						StartTrace = pawn1.Location;
						EndTrace = StartTrace + default.gunAccuracy * (FRand()-0.5)*Y*1000 + default.gunAccuracy * (FRand()-0.5)*Z*1000;
						EndTrace += curMaxRange * X;
						StartTrace.Z += pawn1.BaseEyeHeight;
						EndTrace.Z += pawn1.BaseEyeHeight;
					
						a = Trace(HitLocation, HitNormal, EndTrace, StartTrace, True);
						
						if(a != None)
						{
							a.TakeDamage(GetWeaponDamage(target), Pawn(Owner), HitLocation, vect(4,5,1), dmgType);
							TrySpawnBlood(a, dmgType, HitLocation, HitNormal);
							PlayAppropriateSound(HitLocation, HitNormal, a);
						}
						
						SpawnServerTracer(EndTrace, false);
					}
					
				}
			}
			else
			{
				//projectiles still spawn serverside (they already sync with moving targets correctly)
				ProjectileFire2(ProjectileClass, ProjectileSpeed, bWarnTarget, trackTarget);
			}
			
			curNumShot += 1;
			
			if(curNumShot >= numShots)
			{
				if(AmmoIsDefault() == true)
					nextShotTime = Level.TimeSeconds + shotTime;
				else
					nextShotTime = level.TimeSeconds + 1.5;
			
				//nextShotTime = Level.TimeSeconds + shotTime;
				curNumShot = 0;
			}
		}
		else
		{
			//broadcastmessage("debug: shot was too early. Current time: "$Level.TimeSeconds$", nextShotTime: "$nextShotTime);
			
			//client has fired early. In case of Unreal timing mistakes, queue this shot and fire it asap
			//only store one shot ahead, and only if it's within 100ms
			if(nextShotTime - Level.TimeSeconds <= 0.1)
			{
				qShotTime = nextShotTime;
				qTarget[curNumShot] = target;
				qOffset[curNumShot] = offset;
				
				//broadcastmessage("debug: shot queued. qShotTime: "$qShotTime);
			}
		}
	}
}

function SpawnServerTracer(vector endAt, bool artificialSpread)
{
	local rotator rot;
	local vector startFrom,X,Y,Z;
	
	if(AmmoName == class'AmmoBattery' || AmmoName == Class'DeusEx.AmmoNone')
		return;
	
	startFrom = Owner.Location;
	startFrom.Z += Pawn(Owner).BaseEyeHeight;
	
	if(artificialSpread) //for inaccurate weps *without* a target, fake some spread because we lack dir info
	{
		GetAxes(Pawn(Owner).ViewRotation, X, Y, Z);
		endAt = startFrom + default.gunAccuracy * (FRand()-0.5)*Y*1000 + default.gunAccuracy * (FRand()-0.5)*Z*1000;
		endAt += mpMaxRange * X;
		endAt.Z += Pawn(Owner).BaseEyeHeight;
	}
	
	//we don't want tracers shooting out of our eyes - go a little lower
//	if(Pawn(Owner).bIsCrouching)
//		startFrom.Z -= (Owner.CollisionHeight * 0.25);
//	else
		startFrom.Z -= (Owner.CollisionHeight * 0.35);
	
	rot = Rotator(endAt - startFrom);
	Spawn(Class'TracerServer', Owner,, startFrom, rot);
}

function TrySpawnBlood(actor a, name DamageType, vector HitLocation, vector HitNormal)
{
	if(a.IsA('Pawn') && !a.IsA('Robot'))
	{
		if((DamageType != 'Stunned') && (DamageType != 'Radiation') && (DamageType != 'EMP') && (DamageType != 'NanoVirus') && (DamageType != 'KnockedOut'))
		{
			SpawnBlood(HitLocation, HitNormal);
		}
	}
}

function int GetWeaponDamage(optional actor target)
{
	local int userLevel;
	local float mult;
	local DeusExPlayer player;
	local Pawn p;
	
	player = DeusExPlayer(Owner);
	mult = 1; //for comb str
	
	if(player != None)
	{
		userLevel = player.SkillSystem.GetSkillLevel(GoverningSkill);
		
		if(player.AugmentationSystem.GetClassLevel(class'AugTarget') > 0)
			userLevel += 1;
			
		if(bHandToHand && player.AugmentationSystem.GetClassLevel(class'AugCombat') > 0)
			mult += 0.5;
	}
	else //NPCs
	{
		userLevel = int(Pawn(Owner).Skill) - 1; //DX NPCs will default to skill level 1
	}
	
	return mult * GetWeaponDamageSkill(userLevel, target);
}

function int GetWeaponDamageSkill(int userLevel, optional actor target) //target lets weapons run their own code for falloff, backstab, whatever
{
	//these default values are just an EXAMPLE. Please override them!
	
	if(userLevel == 0) //untrained (sp-only?)
		return 1;
	if(userLevel == 1) //trained
		return 2;
	if(userLevel == 2) //advanced
		return 3;
	if(userLevel == 3) //master
		return 4;
	if(userLevel == 4) //master + targ
		return 5;
}

simulated function float GetWeaponRecoil()
{
	local int userLevel;
	local DeusExPlayer player;
	
	player = DeusExPlayer(Owner);
	
	if(player != None)
	{
		userLevel = player.SkillSystem.GetSkillLevel(GoverningSkill);

		if(player.AugmentationSystem.GetClassLevel(class'AugTarget') > 0)
			userLevel += 1;
			
		return GetWeaponRecoilSkill(userLevel);
	}
}

simulated function float GetWeaponRecoilSkill(float userLevel)
{
	//these default values are just an EXAMPLE. Please override them!
	
	if(userLevel == 0) //untrained (sp-only?)
		return 0.8;
	if(userLevel == 1) //trained
		return 0.5;
	if(userLevel == 2) //advanced
		return 0.2;
	if(userLevel == 3) //master
		return 0;
	if(userLevel == 4) //master + targ
		return 0;
}

function ProjectileFire2(class<projectile> ProjClass, float ProjSpeed, bool bWarn, optional Actor trackTarget) //edited from Engine.u
{
	local Vector Start, X,Y,Z;
	local Pawn PawnOwner;
	local Projectile proj;
	
	PawnOwner = Pawn(Owner);
	Owner.MakeNoise(PawnOwner.SoundDampening);
	GetAxes(PawnOwner.ViewRotation,X,Y,Z);
	Start = Owner.Location + CalcDrawOffset() + FireOffset.X * X + FireOffset.Y * Y + FireOffset.Z * Z;
	AdjustedAim = PawnOwner.AdjustAim(ProjSpeed, Start, AimError, True, bWarn);
	proj = Spawn(ProjClass, Owner,, Start,AdjustedAim); //for some reason, this didn't previously set owner
	
	if(proj != None)
	{
		if(AmmoIsDefault() == true)
			proj.Damage = GetWeaponDamage();

		if(trackTarget != None && bCanTrack)
		{
			DeusExProjectile(proj).Target = trackTarget;
			DeusExProjectile(proj).bTracking = True;
		}
	}
}

function dsoc(int dsocType)
{
	local Pawn aPawn, bPawn;
	
	//if(role == ROLE_AUTHORITY)
	if(Level.NetMode == NM_DedicatedServer)
	{
		if(dsocType == DSOC_FIREANIMS)
		{
			for(aPawn = Level.PawnList; aPawn != None; aPawn = aPawn.nextPawn)
			{
				if(aPawn != Owner)
				{
					bPawn = aPawn;
					
					if(bPawn != None)
					{
						bPawn.Event = 'dsoc1'; //big hack. Can't call ANNA functions directly because ANNAExt must be loaded first
						bPawn.Trigger(self, Pawn(Owner));
					}
				}
			}
		}
		else if(dsocType == DSOC_DEMOSOUND)
		{
			Owner.Event = 'dsoc2';
			Owner.Trigger(self, Pawn(Owner));
		}
	}
}

//
//	SpawnEffects - Spawns the effects like it did in single player
//
simulated function SpawnEffectsClient(Vector HitLocation, Vector HitNormal, Actor Other, float Damage)
{
	local TraceHitSpawner hitspawner;
	local Name damageType;

	damageType = WeaponDamageType();

	if(bPenetrating)
	{
		if (bHandToHand)
			hitspawner = Spawn(class'TraceHitHandSpawner',Other,,HitLocation,Rotator(HitNormal));
		else
			hitspawner = Spawn(class'TraceHitSpawner',Other,,HitLocation,Rotator(HitNormal));
   }
   else
   {
		if (bHandToHand)
			hitspawner = Spawn(class'TraceHitHandNonPenSpawner',Other,,HitLocation,Rotator(HitNormal));
		else
			hitspawner = Spawn(class'TraceHitNonPenSpawner',Other,,HitLocation,Rotator(HitNormal));
	}
   
	if (hitSpawner != None)
	{
		hitspawner.HitDamage = Damage;
		hitSpawner.damageType = damageType;
	}
	
	if (bHandToHand)
	{
		// if we are hand to hand, play an appropriate sound
		if (Other.IsA('DeusExDecoration'))
			Owner.PlaySound(Misc3Sound, SLOT_None,,, 1024);
		else if (Other.IsA('Pawn'))
			Owner.PlaySound(Misc1Sound, SLOT_None,,, 1024);
		else if (Other.IsA('BreakableGlass'))
			Owner.PlaySound(sound'GlassHit1', SLOT_None,,, 1024);
		else if (GetWallMaterial(HitLocation, HitNormal) == 'Glass')
			Owner.PlaySound(sound'BulletProofHit', SLOT_None,,, 1024);
		else
			Owner.PlaySound(Misc2Sound, SLOT_None,,, 1024);
	}
}

simulated function ScopeToggle()
{
	//if(IsInState('Idle')) //going to let people pre-scope, useful for laggy players lol 
	
	if (bHasScope && (Owner != None) && Owner.IsA('DeusExPlayer'))
	{
		if (bZoomed)
			ScopeOff();
		else
			ScopeOn();
	}
}

//copied for future editing
function ScopeOn() //prev not sim 
{
	if (bHasScope && !bZoomed && (Owner != None) && Owner.IsA('DeusExPlayer'))
	{
		// Show the Scope View
		bZoomed = True;
		RefreshScopeDisplay(DeusExPlayer(Owner), False, bZoomed);
	}
}

function ScopeOff()
{
	if (bHasScope && bZoomed && (Owner != None) && Owner.IsA('DeusExPlayer'))
	{
		bZoomed = False;
		// Hide the Scope View
      RefreshScopeDisplay(DeusExPlayer(Owner), False, bZoomed);
		//DeusExRootWindow(DeusExPlayer(Owner).rootWindow).scopeView.DeactivateView();
	}
}

//this is borrowed from Lexichu's DXWeapon2!
simulated function RefreshScopeDisplay(DeusExPlayer player, bool bInstant, bool bScopeOn)
{
	local DeusExRootWindow root;
	local CrateUnbreakableSmall cus;

	if(Player != None)
	{
		root = DeusExRootWindow(Player.RootWindow);
		if(root.scopeView != None)
		{
			if(root.scopeView.Class == Class'DeusEx.DeusExScopeView')
			{
				root.scopeView.Destroy();
				root.scopeView = None;
				root.scopeView = DeusExScopeView(player.RootWindow.NewChild(Class'FixedDXScopeView', False)); //especially this
				root.scopeView.SetWindowAlignments(HALIGN_Full, VALIGN_Full, 0, 0);
			}

			if (bScopeOn)
				root.scopeView.ActivateView(ScopeFOV, False, bInstant);
			else
				root.scopeView.DeactivateView();
		}
	}
}

simulated state SimActive
{
Begin:
	if(bWasInFiring )
	{
		if (bHasMuzzleFlash)
			EraseMuzzleFlashTexture();
		FinishAnim();
	}
	
	bInProcess = False;
	bFiring=False;
	if(bMissingAnims == false) PlayAnim('Select',1.0,0.05);
	//FinishAnim(); //FUCKER. This single line made vanilla wep switching horrible
	SimFinish();
}

simulated state SimIdle
{
	function Timer()
	{
		PlayIdleAnim();
	}
	
Begin:
	bInProcess = False;
	bFiring = False;
	
//	if (!bNearWall)
//		PlayAnim('Idle1',,0.1);

	SetTimer(3.0, True);
}

simulated function TweenDown()
{
	if(bMissingAnims == false) PlayAnim('Down', 2.0, 0.05);
	//nextShotTime = Level.TimeSeconds + 10000;
	//SetClientCanFireTime(10000);
}

function ReloadAmmo() //wep anim times: https://pastebin.com/Hg2Q4XpT
{
	if((ReloadCount == 0) || (bHandToHand == True)) //single use or hand to hand weapon if ReloadCount == 0
	{
		Pawn(Owner).ClientMessage(msgCannotBeReloaded);
		return;
	}
	
	clipcount = reloadcount; //set clip to 0
	//broadcastmessage("state is currently: "$GetStateName());

	if((bIsReloading == false) && (CheckForAmmo(true)) == true)
		GotoState('ANNAReload');
}

state ANNAReload
{
Begin:

	nextShotTime = nextShotTime + 10000; //these two lines suck... I need to learn sim functions properly lol
	SetClientCanFireTime(10000); //don't fire while reloading (hack)
	
	bWasZoomed = false;
	Sleep(0.15); //let weps play a fire anim before reloading

	if(bHasScope && (Owner != None) && Owner.IsA('DeusExPlayer'))
	{
		if(bZoomed)
		{
		//	Sleep(0.15); //let players see the result of their shot before zooming out
			bWasZoomed = true;
			ScopeOff();
		}
	}

	//these anim timings are serverside (I think), so they can't be sped up by clients. Will switch to ms timing if this is exploitable
	bIsReloading = true; //can probably just check state instead of using this var
	PlaySound(CockingSound, SLOT_None, TransientSoundVolume);
	if(bMissingAnims == false) PlayAnim('ReloadBegin', 1.0, 0.15); //last param is tween-time
	FinishAnim();
	if(bMissingAnims == false) PlayAnim('Reload', 1.0, 0.2);
	FinishAnim();
	clipcount = 0;
	if(bMissingAnims == false) PlayAnim('ReloadEnd', 1.0, 0.15);
	PlaySound(AltFireSound, SLOT_None, TransientSoundVolume);
	FinishAnim();
	Sleep(0.1); //hm
	
	if(bHasScope && (Owner != None) && Owner.IsA('DeusExPlayer'))
	{
		if(bWasZoomed)
			ScopeOn();
	}
	
	nextShotTime = 0;
	SetClientCanFireTime(0);
	bIsReloading = false;
	GotoState('Idle');
}

simulated function bool AmmoIsDefault()
{
	//Pawn(Owner).ClientMessage("ammotype: "$AmmoType.Class$", ammoname:"$AmmoNames[0]);
	
	if(AmmoNames[1] != None)
	{
		if(AmmoType.Class != AmmoNames[0])
			return false;
	}
	
	return true;
}

simulated function SetClientCanFireTime(float canFireTime)
{
	//was using Level.TimeSeconds as a param calling this, but server and client have different Level.TimeSeconds! So it was causing weird client nst values
	
	if(canFireTime == 0)
		nextShotTime = 0;
	else
		nextShotTime += canFireTime;
}

//currently unused. todo: rework the stupid muzzleflash system. Atm, this works fine but the light doesn't 'stick' to the player for long?
simulated function MuzzleFlashLight2(optional float flashDuration)
{
	local Vector offset, X, Y, Z;

 	if (!bHasMuzzleFlash)
		return;

	if ((flash != None) && !flash.bDeleteMe)
		flash.LifeSpan = flash.Default.LifeSpan;
	else
	{
		GetAxes(Pawn(Owner).ViewRotation,X,Y,Z);
		offset = Owner.Location;
		offset += X * Owner.CollisionRadius * 2;
		flash = spawn(class'MuzzleFlash',,, offset);
		flash.SetBase(Owner);
		
		if(flashDuration > 0)
			flash.LifeSpan = flashDuration;
		else
			flash.LifeSpan = flash.Default.LifeSpan;
	}
}

function CycleAmmo()
{
//	testVar = !testVar;
//	broadcastmessage("debug: using old code: "$testVar);

	Super.CycleAmmo();
	//clipcount = reloadcount;
}

//for some reason, some weapons would not give full ammo when frobbed only once, so force max ammo every time
/*
auto state Pickup
{
	function Frob (Actor Frobber, Inventory frobWith)
	{
		local Ammo AT;
		local DeusExPlayer P;
		
		Super.Frob(Frobber, frobWith);
		
		if(AmmoType == None)
			return;
			
		if(bTossedOut)
			return;
			
		P = DeusExPlayer(Frobber);
		
		if(P != None)
		{
			if (AmmoNames[0] == None)
				AT = Ammo(P.FindInventoryType(AmmoName));
			else
				AT = Ammo(P.FindInventoryType(AmmoNames[0]));
				
			if (AT != None)
				AT.AmmoAmount = AT.MaxAmmo;
		}
	}
}*/

//the original function. iirc, there were conditions where it could cause a crash (inf loop)
/* 
function CycleAmmo()
{
	local int i, last;

	if (NumAmmoTypesAvailable() < 2)
		return;

	for (i=0; i<ArrayCount(AmmoNames); i++)
		if (AmmoNames[i] == AmmoName)
			break;

	last = i;

	do
	{
		if (++i >= 3)
			i = 0;

		if (LoadAmmo(i))
			break;
	} until (last == i);
}
*/

//just in case we want to edit this later
/*
function name WeaponDamageType()
{
	local name                    damageType;
	local Class<DeusExProjectile> projClass;

	projClass = Class<DeusExProjectile>(ProjectileClass);
	if (bInstantHit)
	{
		if (StunDuration > 0)
			damageType = 'Stunned';
		else
			damageType = 'Shot';

		if (AmmoType != None)
			if (AmmoType.IsA('AmmoSabot'))
				damageType = 'Sabot';
	}
	else if (projClass != None)
		damageType = projClass.Default.damageType;
	else
		damageType = 'None';

	return (damageType);
}
*/

//-------------------------------------------------------------------
// Hit markers. Stuff below was ported from Lexichu's DXW2 mod
//-------------------------------------------------------------------
simulated function UpdateHitMarker(Actor Other, vector HitLocation)
{}
/*
simulated function UpdateHitMarker(Actor Other, vector HitLocation)
{
	local DeusExPlayer	dxPlayer, dxTarg;
	local ScriptedPawn	pTarg;
	local vector		offsetXYZ;
	local Sound			hitConfirmSFX;

	dxPlayer = DeusExPlayer(Owner);
	dxTarg = DeusExPlayer(Other);
	pTarg = ScriptedPawn(Other);

	if(dxPlayer == None)
		return;

	if(Role != ROLE_Authority)
		return;

	if(dxTarg != None)
	{
		bHeadshotMarker = (dxTarg.GetMPHitLocation(HitLocation) == 1);
	}
	else if(pTarg != None)
	{
		if(!pTarg.IsA('Robot') && !pTarg.IsA('Animal'))
		{
			offsetXYZ = (hitLocation - pTarg.Location) << pTarg.Rotation;
			bHeadshotMarker = NPCHeadshotCheck(offsetXYZ, pTarg, False);
		}
		else
		{
			bHeadshotMarker = False; // Can't headshot that without a head!
		}
	}

	HMTex = Texture(DynamicLoadObject(HitM_Texture, class'Texture'));
	FadeOutSeconds = 0.2;

	if((bHeadshotMarker && !bHSS) || !bSS)
	{
		if(bHeadshotMarker)
			HitConfirmSFX = Sound(DynamicLoadObject(HitM_Head, class'Sound'));
		else
			HitConfirmSFX = Sound(DynamicLoadObject(HitM_Normal, class'Sound'));
		dxPlayer.ClientPlaySound(HitConfirmSFX);
	}

	bHSS = bHeadshotMarker;
	bSS = True;
}*/

function bool NPCHeadshotCheck(vector Offset, ScriptedPawn victim, optional bool bSidesAreHead)
{
	local float headOffsetZ, headOffsetY;

	headOffsetZ = victim.CollisionHeight * 0.7;
	headOffsetY = victim.CollisionRadius * 0.3;

	if (offset.z > headOffsetZ)		// head
	{
		// narrow the head region
		if ((Abs(offset.x) < headOffsetY) || (Abs(offset.y) < headOffsetY))
		{
			return True;
		}
		else  // sides of head treated as torso
		{
			return bSidesAreHead;
		}
	}

	return False;
}

/*
simulated event RenderOverlays( canvas Can )
{
	local int MarkerBrightness;
	local pawn p;
	local DeusExPlayer DXP;

	Super.RenderOverlays(Can);
	DXP = DeusExPlayer(Owner);

	if(DXP != None)
	{
		if(HMTex != None)
		{
			if(FadeOutSeconds < 0.0)
				FadeOutSeconds = 0.0;

			MarkerBrightness = FadeOutSeconds * 1275;
			Can.Style = 3;
			
			if(bHeadshotMarker || bHSS)
			{
				Can.SetPos((Can.SizeX / 2) - HMTex.uClamp * 0.75, (Can.SizeY / 2) - HMTex.vClamp * 0.75);
				Can.DrawColor.R = MarkerBrightness;
				Can.DrawColor.G = 0;
				Can.DrawColor.B = 0;
				Can.DrawTile(HMTex, HMTex.uClamp * 1.5, HMTex.vClamp * 1.5, 0, 0, HMTex.uClamp, HMTex.vClamp);
			}
			else
			{
				Can.SetPos((Can.SizeX / 2) - HMTex.uClamp * 0.5, (Can.SizeY / 2) - HMTex.vClamp * 0.5);
				Can.DrawColor.R = MarkerBrightness;
				Can.DrawColor.G = MarkerBrightness;
				Can.DrawColor.B = MarkerBrightness;
				Can.DrawTile(HMTex, HMTex.uClamp, HMTex.vClamp, 0, 0, HMTex.uClamp, HMTex.vClamp);
			}
		}
	}
}
*/

//------------------
// Pruned functions
//------------------
function Fire(float F)
{
	local ScriptedPawn p;
	local int i;
	
	p = ScriptedPawn(Owner);
	
	if(p!=None)
	{
		if(burstDuration > 0 && Level.TimeSeconds >= stopBurstTime)
			stopBurstTime = Level.TimeSeconds + burstDuration;
						
		for(i=0; i < numShots; i++)
			FireOnServer();
	}
}

function ServerHandleNotify( bool bInstantHit, class<projectile> ProjClass, float ProjSpeed, bool bWarn ){}
function AltFire(float F){}
function Reload(){}
function Projectile ProjectileFire(class<projectile> ProjClass, float ProjSpeed, bool bWarn){}
simulated function bool ClientFire(float Value){return false;}
simulated function bool ClientAltFire(float Value){return false;}
simulated function HandToHandAttack(){}
simulated function TraceFire( float Accuracy ){}

defaultproperties
{
	HitFeedback=Sound'CDXExtension.Generic.QFeedBack'
	recoilDuration=0.15
	altAmmoDelayTime=1.0
	shotTime=1.0
	atkDelayTime=0.25
	deleteAfterFireTime=0
	delWepTime=0
	lowClipCount=2
    ScopeFOV=20
	gunAccuracy=0
	numShots=1
	bringUpTime=0.5
	bIsReloading=False
	bInstantHit=True
}